﻿
using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Web.UI;
using System.Web.UI.WebControls;
using Microsoft.IdentityModel.Clients.ActiveDirectory;
using System.Threading.Tasks;
using System.Net.Http;
using System.Collections.Specialized;
using Newtonsoft.Json;
using System.Web.Security;
using System.Net;
using System.IO;


namespace PBIWebApp
{
 /* NOTE: This sample is to illustrate how to authenticate a Power BI web app. 
 * In a production application, you should provide appropriate exception handling and refactor authentication settings into 
 * a configuration. Authentication settings are hard-coded in the sample to make it easier to follow the flow of authentication. */
    public partial class _Default : Page
    {
        public AuthenticationResult authResult { get; set; }
        string baseUri = "https://api.powerbi.com/beta/myorg/";

        protected void Page_Load(object sender, EventArgs e)
        {

            if (Session["authResult"] != null)
            {
                authResult = (AuthenticationResult)Session["authResult"];
                accessTokenTextbox.Text = authResult.AccessToken;
                return;
            }

			try
			{

                bool bypassPowerBILogin = Properties.Settings.Default.BypassPowerBILogin;
                if (bypassPowerBILogin)
                {
                // As it stands, Power BI doesn't support custom security so the user must be registered with Power BI
                // After you authenticate the user, you need to retrieve the Power BI user credentials 
                string powerBIUserID = Properties.Settings.Default.PowerBIUserID;
                System.Security.SecureString powerBIPassword = Util.DecryptString(Properties.Settings.Default.PowerBIPassword);

                // perform two-leg OAuth
                   // update tenantID in web.config file as instructed in https://support.office.com/en-us/article/Find-your-Office-365-tenant-ID-6891b561-a52d-4ade-9f39-b492285e2c9b
                   System.Net.WebRequest request = System.Net.WebRequest.Create(String.Format("https://login.microsoftonline.com/{0}/oauth2/token", Properties.Settings.Default.TenantID)) as System.Net.HttpWebRequest;
                   request.ContentType = "application/x-www-form-urlencoded";
                   request.Method = WebRequestMethods.Http.Post;
                   using (StreamWriter streamWriter = new StreamWriter(request.GetRequestStream()))
                   {
                       string payload = String.Format("grant_type=password&client_id={0}&client_secret={1}&resource=https%3a%2f%2fanalysis.windows.net%2fpowerbi%2fapi&username={2}&password={3}", 
                           WebUtility.UrlEncode(Properties.Settings.Default.ClientID), WebUtility.UrlEncode(Properties.Settings.Default.ClientSecret),
                           WebUtility.UrlEncode(powerBIUserID), WebUtility.UrlEncode(Util.ToInsecureString(powerBIPassword)));
                       streamWriter.Write(payload);
                   }
                   using (HttpWebResponse response = (HttpWebResponse)request.GetResponse())
                   {
                       using (StreamReader streamReader = new StreamReader(response.GetResponseStream()))
                       {
                           string payload = streamReader.ReadToEnd();
                           payload = payload.Replace("access_token", "AccessToken");
                           payload = payload.Replace("refresh_token", "RefreshToken");

                           AuthenticationResult ar = AuthenticationResult.Deserialize(payload);
                           Session["authResult"] = ar;
                           accessTokenTextbox.Text = ar.AccessToken;
                       }
                   }
                }
                else
                {
                    //perform three-leg OAuth
                    string authorityUri = "https://login.windows.net/common/oauth2/authorize/";
                    //Create a query string
                    //Create a sign-in NameValueCollection for query string
                    var @params = new NameValueCollection
                    {
                        //Azure AD will return an authorization code. 
                        //See the Redirect class to see how "code" is used to AcquireTokenByAuthorizationCode
                        {"response_type", "code"},

                        //Client ID is used by the application to identify themselves to the users that they are requesting permissions from. 
                        //You get the client id when you register your Azure app.
                        {"client_id", Properties.Settings.Default.ClientID},

                        //Resource uri to the Power BI resource to be authorized
                        {"resource", "https://analysis.windows.net/powerbi/api"},

                        //After user authenticates, Azure AD will redirect back to the web app http://localhost:13528/Redirect
                        {"redirect_uri", String.Format("{0}://{1}/Redirect", HttpContext.Current.Request.Url.Scheme, 
                            HttpContext.Current.Request.Url.Authority)}
                    };
                    //Create sign-in query string
                    var queryString = HttpUtility.ParseQueryString(string.Empty);
                    queryString.Add(@params);
                    Session["redirectUrl"] = "/Reports/Report.aspx";
                    Response.Redirect(String.Format("{0}?{1}", authorityUri, queryString));   
                }

			}
			catch(Exception ex)
			{
				userLabel.Text = ex.Message;
				return;
			}
        }

    

        protected void getReportsButton_Click(object sender, EventArgs e)
        {
            string responseContent = string.Empty;
            //Configure datasets request
            System.Net.WebRequest request = System.Net.WebRequest.Create(String.Format("{0}reports", baseUri)) 
                as System.Net.HttpWebRequest;
            request.Method = "GET";
            request.ContentLength = 0;
            request.Headers.Add("Authorization", String.Format("Bearer {0}", authResult.AccessToken));

            //Get datasets response from request.GetResponse()
            using (var response = request.GetResponse() as System.Net.HttpWebResponse)
            {
                //Get reader from response stream
                using (var reader = new System.IO.StreamReader(response.GetResponseStream()))
                {
                    responseContent = reader.ReadToEnd();

                    //Deserialize JSON string
                    PBIReports PBIReports = JsonConvert.DeserializeObject<PBIReports>(responseContent);

                    tb_reportsResult.Text = string.Empty;
                    //Get each Dataset from 
                    foreach (PBIReport rpt in PBIReports.value)
                    {
                        //https://wabi-staging-us-east-redirect.analysis.windows.net
                        tb_reportsResult.Text += String.Format("{0}\t{1}\t{2}\n", rpt.id, rpt.name, 
                            "https://app.powerbi.com/reportEmbed?reportId="+rpt.id);

                        //tb_reportsResult.Text += String.Format("{0}\t{1}\t{2}\n", rpt.id, rpt.name, rpt.embedUrl); -- [TL] as of now, the link doesn't work because embedUrl uses embedReport while it should be reportEmbed
                    }
                }
            }
        }
    }
    
  
    public class PBIReports
    {
        public PBIReport[] value { get; set; }
    }
    public class PBIReport
    {
        public string id { get; set; }

        // the name of this property will change to 'displayName' when the API moves from Beta to V1 namespace
        public string name { get; set; }

        public string webUrl { get; set; }

        public string embedUrl { get; set; } 

    }




}